# Autores: Laura Medina Henche y Emanuele Schiavi.
# Universidad Rey Juan Carlos, 2024.
# El presente software se distribuye segun la licencia
# GPLv3 (https://www.gnu.org/licenses/gpl-3.0.en.html)


from sympy import *
from sympy.plotting import plot3d, PlotGrid

a, t, T, x0, i, k = symbols('a t T x0 i k')
b = Symbol('b', nonzero=True, noninfinite=True)

# Definimos los parametros
p = [a, b]

# Declaramos las funciones que utilizaremos para la EDO
x1 = Function('x1')
dx1 = x1(t).diff(t)

x2 = Function('x2')
dx2 = x2(t).diff(t)

x = Matrix([x1(t), x2(t)])
dx = Matrix(x.diff(t))
dxp = Matrix([diff(x, v) for v in p])

y1 = sin(t)
dy1 = diff(y1, t)

y2 = cos(t)
dy2 = diff(y2, t)

y = Matrix([y1, y2])
dy = diff(y, t)
dyp = Matrix([diff(y, v) for v in p])

# Definimos la funcion lambda para el adjunto
lmbd1 = Function('lambda1')
lmbd2 = Function('lambda2')
lmbd = Matrix([lmbd1(t), lmbd2(t)])
lmbdT = Matrix([lmbd[i].subs({t: T}) for i in range(len(lmbd))])

# Fijamos la funcion f(x,p,t) a integrar
lista = [sympify(Pow((x[i] - y[i]), 2), Function) for i in range(len(x))]
suma = 0
for i in range(len(lista)):
    suma = suma + lista[i]
f = (1/2) * suma
print("f= ", f)

# Calculamos las derivadas parciales de f
dfx = diff(f, x)
print("dfx= ", dfx)
dfp = Matrix([diff(f, v) for v in p])
print("dfp= ", dfp)

# Fijamos la restriccion explicita dada por la EDO
h_exp = Matrix([b * fun for fun in x])
print("h_exp= ", h_exp)

# Fijamos la EDO implicita
h = dx - h_exp
print("h= ", h)

# Calculamos las parciales
dhx = []
dhdx = []
dhp = []
for i in range(len(h)):
    dhx.append([diff(h[i], v) for v in x])
    dhdx.append([diff(h[i], v) for v in dx])
    dhp.append([diff(h[i], v) for v in p])
dhx = Matrix(dhx)
dhdx = Matrix(dhdx)
dhp = Matrix(dhp)
print("dhx= ", dhx)
print("dhdx= ", dhdx)
print("dhp= ", dhp)

# Fijamos la funcion g de restricciones
x0 = Matrix([x[i].subs({t: 0}) for i in range(len(x))])
print("x0= ", x0)
g = Matrix([x0[i] - a for i in range(len(x0))])
print("g= ", g)

# Calculamos sus parciales
dgx0 = []
dgp = []
for i in range(len(h)):
    dgx0.append([diff(g[i], v) for v in x0])
    dgp.append([diff(g[i], v) for v in p])
dgx0 = Matrix(dgx0)
dgp = Matrix(dgp)
print("dgx0= ", dgx0)
print("dgp= ", dgp)

# Ahora que tenemos las primeras variables definidas,
# resolvemos la primera EDO, la principal
ics = [{x0[i]: a} for i in range(len(x0))]
resODE = Matrix([simplify(dsolve(h[i], ics=ics[i])).rhs for i in range(len(h))])
resODE = Matrix([sympify(resODE[i], Function) for i in range(len(resODE))])
print("Resultado primera EDO: ", resODE)

# Vamos al segundo paso del algoritmo
ics = [{lmbdT[i]: 0} for i in range(len(lmbdT))]
EDO_adj = dfx.subs({x[n]: resODE[n] for n in range(len(resODE))}).T + lmbd.T * (dhx - dhdx.diff(t)) - \
          lmbd.diff(t).T * dhdx
print("EDO_adj: ", EDO_adj)
res_adj = Matrix([sympify(factor(simplify(dsolve(EDO_adj[i], ics=ics[i])).rhs), Function) for i in range(len(resODE))])
print("Resultado ecuacion del adjunto: ", res_adj)


# Tercer paso del algoritmo
dfp = dfp.T
dFpCal = []
for k in range(len(p)):
    f_int = dfp[:, k] + res_adj.T * dhp[:, k]
    f_int = f_int.subs({x[k]: resODE[k] for k in range(len(resODE))})
    integral = integrate(f_int, (t, 0, T), conds='none')
    Fp = integral + res_adj.T.subs(t, 0) * dhdx.subs(t, 0) * dgx0**-1 * dgp[:, k]
    dFpCal.append(factor(simplify(factor(Fp))))

dFpCal = Matrix(dFpCal)
print("DFPCal= ", dFpCal)


# Calculo directo del funcional
Fp = []
print("Calculamos la integral directamente: ")
Fun = integrate(f.subs({x[k]: resODE[k] for k in range(len(x))}), (t, 0, T), conds='none')
Fun = sympify(simplify(Fun), Function)
print("Fp = ", Fun)
print("Comprobamos que el resultado es correcto...")
for v in p:
    dif = simplify(diff(Fun, v))
    Fp.append(dif)
    print("dFp", v, " directo= ", dif)

# Creamos nuestros graficos
t1 = "Integral directa"
plot3d(Fp[i].subs({T: 2}), (p[0], 0.001, 2), (p[1], 0.001, 2), title=t1, xlabel=p[0], ylabel=p[1])

t2 = "Primer componente del\ngradiente"
pa = plot3d(dFpCal[0].subs({T: 2}), (p[0], 0.001, 2), (p[1], 0.001, 2), title=t2, xlabel=p[0], ylabel=p[1],
            show=False)
t3 = "Segundo componente del\ngradiente"
pb = plot3d(dFpCal[1].subs({T: 2}), (p[0], 0.001, 2), (p[1], 0.001, 2), title=t3, xlabel=p[0], ylabel=p[1],
            show=False)
PlotGrid(1, 2, pa, pb, margin=0.7)

t4 = "Modulo del gradiente"
modulo = sqrt((dFpCal[0].subs({T: 2})) ** 2 + (dFpCal[1].subs({T: 2})) ** 2)
plot3d(modulo, (p[0], 0.001, 2), (p[1], 0.001, 2), title=t4, xlabel=p[0], ylabel=p[1])

